/*
This file is part of uvcc -- USB Video Class (UVC) device configurator.
Copyright (C) 2018, 2019, 2020, 2021 Joel Purra <https://joelpurra.com/>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
*/
import arrayNonUniq from "array-non-uniq";
import filterObject from "filter-obj";
import assert from "node:assert";
import sortKeys from "sort-keys";
import WrappedError from "./utilities/wrapped-error.js";
/* eslint-disable @typescript-eslint/member-ordering */
export default class CameraControlHelper {
    constructor(
    // eslint-disable-next-line @typescript-eslint/naming-convention
    UvcControl, camera) {
        Object.defineProperty(this, "UvcControl", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: UvcControl
        });
        Object.defineProperty(this, "camera", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: camera
        });
        Object.defineProperty(this, "cachedMappedSupportedControls", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: null
        });
        assert.strictEqual(arguments.length, 2);
        assert(typeof this.UvcControl === "function");
        assert(typeof this.camera === "object");
    }
    async getControlNames() {
        const controls = await this.getSupportedControls();
        return Object.keys(controls);
    }
    async getGettableControlNames() {
        const gettableControls = await this.getGettableControls();
        return Object.keys(gettableControls);
    }
    async getRangedControlNames() {
        const rangedControls = await this.getRangedControls();
        return Object.keys(rangedControls);
    }
    async getSettableControlNames() {
        const settableControls = await this.getSettableControls();
        return Object.keys(settableControls);
    }
    isGettableControl(control) {
        // NOTE: relies on uvc-control internals.
        return control.requests.includes(this.UvcControl.REQUEST.GET_CUR);
    }
    isRangedControl(control) {
        // NOTE: relies on uvc-control internals.
        return control.requests.includes(this.UvcControl.REQUEST.GET_MIN)
            && control.requests.includes(this.UvcControl.REQUEST.GET_MAX);
    }
    isSettableControl(control) {
        // NOTE: relies on uvc-control internals.
        return control.requests.includes(this.UvcControl.REQUEST.SET_CUR)
            || (
            // TODO: treat optionally settable controls separately?
            Array.isArray(control.optional_requests)
                && control.optional_requests.includes(this.UvcControl.REQUEST.SET_CUR));
    }
    async mapSupportedControls() {
        const supportedControls = this.camera.supportedControls
            .map((supportedControlName) => this.UvcControl.controls[supportedControlName])
            // TODO: proper ducktyping function.
            .filter((t) => Boolean(t));
        const missingControls = supportedControls.filter((control) => !control);
        if (missingControls.length > 0) {
            throw new Error(`Controls were not found: ${JSON.stringify(missingControls)}`);
        }
        const nonUniqueControls = arrayNonUniq(supportedControls);
        if (nonUniqueControls.length > 0) {
            throw new Error(`Controls were already found: ${JSON.stringify(nonUniqueControls)}`);
        }
        const mappedControls = supportedControls.map((control) => {
            try {
                const uvccControl = [
                    control.name,
                    {
                        // NOTE: minimal uvcc-specific mapping.
                        isGettable: this.isGettableControl(control),
                        isRanged: this.isRangedControl(control),
                        isSettable: this.isSettableControl(control),
                    },
                ];
                return uvccControl;
            }
            catch (error) {
                if (error instanceof Error) {
                    const wrappedError = new WrappedError(error, `Could not map control: ${JSON.stringify(control.name)}`);
                    throw wrappedError;
                }
                throw error;
            }
        });
        const controlsObject = {};
        for (const mappedControl of mappedControls) {
            controlsObject[mappedControl[0]] = mappedControl[1];
        }
        const sortedControls = sortKeys(controlsObject);
        return sortedControls;
    }
    async getSupportedControls() {
        if (this.cachedMappedSupportedControls === null) {
            this.cachedMappedSupportedControls = await this.mapSupportedControls();
        }
        return this.cachedMappedSupportedControls;
    }
    async getGettableControls() {
        const controls = await this.getSupportedControls();
        return filterObject(controls, (_key, control) => control.isGettable);
    }
    async getRangedControls() {
        const controls = await this.getSupportedControls();
        return filterObject(controls, (_key, control) => control.isRanged);
    }
    async getSettableControls() {
        const controls = await this.getSupportedControls();
        return filterObject(controls, (_key, control) => control.isSettable);
    }
}
/* eslint-enable @typescript-eslint/member-ordering */
